/*
 *
 *	RADIUS
 *	Remote Authentication Dial In User Service
 *
 *
 *	Livingston Enterprises, Inc.
 *	6920 Koll Center Parkway
 *	Pleasanton, CA   94566
 *
 *	Copyright 1992 Livingston Enterprises, Inc.
 *
 *	Permission to use, copy, modify, and distribute this software for any
 *	purpose and without fee is hereby granted, provided that this
 *	copyright and permission notice appear on all copies and supporting
 *	documentation, the name of Livingston Enterprises, Inc. not be used
 *	in advertising or publicity pertaining to distribution of the
 *	program without specific prior permission, and notice be given
 *	in supporting documentation that copying and distribution is by
 *	permission of Livingston Enterprises, Inc.   
 *
 *	Livingston Enterprises, Inc. makes no representations about
 *	the suitability of this software for any purpose.  It is
 *	provided "as is" without express or implied warranty.
 *
 */

/* Squid_rad_auth is a RADIUS authenticator for Squid (2.x but it probably
 * can work with 1.1.x and the right patches. I only tested with Squid-2.1.x.
 * The authenticator reads a line with a user and password combination.
 * If access is granted OK is returned. Else ERR.
 *
 * The configuration file is defined in SQUID_RAD_CONF. See conf.h.
 * Squid_rad_auth-1.0 is based on modules from the Cistron-radiusd-1.5.4.
 * There might be some dead code still here. You are invited to clean it
 * up.
 *
 * Currently you should only start 1 authentificator at a time because the
 * the ID's of the different programs can start to conflict. I'm not sure it
 * would help anyway. I think the RADIUS server is close by and I don't think
 * it will handle requests in parallel anyway (correct me if I'm wrong here)
 *
 * Marc van Selm <selm@cistron.nl>
 *
 * $Id: squid_rad_auth.c,v 1.1 1999/01/03 14:40:43 selm Exp $
 */

#include	<sys/types.h>
#include	<sys/socket.h>
#include	<netinet/in.h>

#include	<stdio.h>
#include	<unistd.h>
#include	<netdb.h>
#include	<pwd.h>
#include	<stdlib.h>

#include	"radiusd.h"

#define MAXPWNAM	254
#define MAXPASS		254
#define MAXLINE         254

int		i_send_buffer[2048];
int		i_recv_buffer[2048];
char		*send_buffer = (char *)i_send_buffer;
char		*recv_buffer = (char *)i_recv_buffer;
char		*progname = "squid_rad_auth";
int		sockfd;
char		vector[AUTH_VECTOR_LEN];
char		secretkey[MAXPASS+1];

char		*radius_dir = RADIUS_DIR;
char		*radlog_dir = NULL;
int		debug_flag = 0;

/*
 *	Receive UDP client requests, build an authorization request
 *	structure, and attach attribute-value pairs contained in
 *	the request to the new structure.
 */
static AUTH_REQ	*squid_radauth_radrecv(UINT4 host, u_short udp_port,
			char *buffer, int length)
{
	u_char		*ptr;
	AUTH_HDR	*auth;
	int		totallen;
	int		attribute;
	int		attrlen;
	DICT_ATTR	*attr;
	UINT4		lvalue;
	VALUE_PAIR	*first_pair;
	VALUE_PAIR	*prev;
	VALUE_PAIR	*pair;
	AUTH_REQ	*authreq;

	/*
	 *	Pre-allocate the new request data structure
	 */

	if((authreq = (AUTH_REQ *)malloc(sizeof(AUTH_REQ))) ==
						(AUTH_REQ *)NULL) {
		fprintf(stderr, "%s: no memory\n", progname);
		exit(1);
	}

	auth = (AUTH_HDR *)buffer;
	totallen = ntohs(auth->length);

	/*
	printf("squid_radauth_radrecv: Request from host %lx code=%d, id=%d, length=%d\n",
			(u_long)host, auth->code, auth->id, totallen);
	*/

	/*
	 *	Fill header fields
	 */
	authreq->ipaddr = host;
	authreq->udp_port = udp_port;
	authreq->id = auth->id;
	authreq->code = auth->code;
	memcpy(authreq->vector, auth->vector, AUTH_VECTOR_LEN);

	/*
	 *	Extract attribute-value pairs
	 */
	ptr = (u_char *)auth->data;
	length -= AUTH_HDR_LEN;
	first_pair = (VALUE_PAIR *)NULL;
	prev = (VALUE_PAIR *)NULL;

	while(length > 0) {

		attribute = *ptr++;
		attrlen = *ptr++;
		if(attrlen < 2) {
			length = 0;
			continue;
		}
		attrlen -= 2;
		if((attr = dict_attrget(attribute)) == (DICT_ATTR *)NULL) {
			printf("Received unknown attribute %d\n", attribute);
		}
		else if ( attrlen >= AUTH_STRING_LEN ) {
			printf("attribute %d too long, %d >= %d\n", attribute,
				attrlen, AUTH_STRING_LEN);
		}
		else {
			if((pair = (VALUE_PAIR *)malloc(sizeof(VALUE_PAIR))) ==
						(VALUE_PAIR *)NULL) {
				fprintf(stderr, "%s: no memory\n",
						progname);
				exit(1);
			}
			strcpy(pair->name, attr->name);
			pair->attribute = attr->value;
			pair->type = attr->type;
			pair->next = (VALUE_PAIR *)NULL;

			switch(attr->type) {

			case PW_TYPE_STRING:
				memcpy(pair->strvalue, ptr, attrlen);
				pair->strvalue[attrlen] = '\0';
				pair->length = attrlen;
				if(first_pair == (VALUE_PAIR *)NULL) {
					first_pair = pair;
				}
				else {
					prev->next = pair;
				}
				prev = pair;
				break;
			
			case PW_TYPE_INTEGER:
			case PW_TYPE_IPADDR:
				memcpy(&lvalue, ptr, sizeof(UINT4));
				pair->lvalue = ntohl(lvalue);
				if(first_pair == (VALUE_PAIR *)NULL) {
					first_pair = pair;
				}
				else {
					prev->next = pair;
				}
				prev = pair;
				break;
			
			default:
				printf("    %s (Unknown Type %d)\n", attr->name,attr->type);
				free(pair);
				break;
			}

		}
		ptr += attrlen;
		length -= attrlen + 2;
	}
	authreq->request = first_pair;
	return(authreq);
}


/*
 *	Receive and print the result.
 */
int result_recv(UINT4 host, u_short udp_port, char *buffer, int length)
{
	AUTH_HDR	*auth;
	int		totallen;
	char		reply_digest[AUTH_VECTOR_LEN];
	char		calc_digest[AUTH_VECTOR_LEN];
	int		secretlen;
	AUTH_REQ	*authreq;
	VALUE_PAIR	*req;

	auth = (AUTH_HDR *)buffer;
	totallen = ntohs(auth->length);

	if(totallen != length) {
		fprintf(stderr,
			"Squid_rad_auth: Received invalid reply length from server (want %d/ got %d)\n",
			totallen, length);
		return -1;
	}

	/* Verify the reply digest */
	memcpy(reply_digest, auth->vector, AUTH_VECTOR_LEN);
	memcpy(auth->vector, vector, AUTH_VECTOR_LEN);
	secretlen = strlen(secretkey);
	memcpy(buffer + length, secretkey, secretlen);
	md5_calc(calc_digest, (char *)auth, length + secretlen);

	if(memcmp(reply_digest, calc_digest, AUTH_VECTOR_LEN) != 0) {
		printf("Warning: Received invalid reply digest from server\n");
	}

	authreq = squid_radauth_radrecv(host, udp_port, buffer, length);

	/*
	* This prints the radius response, but it confuses the
	* squid authenticator. Don't uncomment them.
	*
	*/
	/*
	req = authreq->request;

	while(req) {
		printf("    ");
		fprint_attr_val(stdout, req);
		printf("\n");
		req = req->next;
	}
	*/

	if(auth->code != PW_AUTHENTICATION_ACK)
		return -1;

	return 0;
}


/*
 *	Generate a random vector.
 */
static void random_vector(char *vector)
{
	int	randno;
	int	i;

	srand(time(0));
	for(i = 0;i < AUTH_VECTOR_LEN;) {
		randno = rand();
		memcpy(vector, &randno, sizeof(int));
		vector += sizeof(int);
		i += sizeof(int);
	}
}

/* read the config file defined by SQUID_RAD_CONF.
 * The format should be something like:
 * # squid_rad_auth configuration file
 * # MvS: 28-10-1998
 * server suncone.cistron.nl
 * secret testje
 */
int rad_auth_config(char *server, char *secretkey)
{	FILE *cf;
	char line[MAXLINE];
	int srv=0, crt=0;
	
	if ( (cf=fopen(SQUID_RAD_CONF, "r"))==NULL)
	{	perror(SQUID_RAD_CONF);
		return -1;
	}
	while ( fgets(line, MAXLINE, cf) != NULL )
	{	if ( !memcmp(line, "server", 6) )
			srv=sscanf(line, "server %s", server);
		if ( !memcmp(line, "secret", 6) )
			crt=sscanf(line, "secret %s", secretkey);
		if ( srv && crt )
			return 0;
	}
	return -1;
}

int main(int argc, char **argv)
{
	int			salen;
	int			result;
	struct	sockaddr	salocal;
	struct	sockaddr	saremote;
	struct	sockaddr_in	*sin;
	struct	servent		*svp;
	struct timeval		tv;
	fd_set			readfds;
        u_short                 svc_port;
 	char			username[MAXPWNAM];
 	char			passwd[MAXPASS];
	char			server[MAXLINE];
 	AUTH_HDR		*auth;
	char			passbuf[MAXPASS];
	char			md5buf[256];
	char			nasname[MAXLINE];
	UINT4			nas_ipaddr;
	UINT4			auth_ipaddr;
	UINT4			ui;
	u_short			local_port;
	u_short			total_length;
	int			portno;
	char			*ptr;
	int			length;
	int			secretlen;
	int			i, j;
	char			authstring[MAXLINE];
	char                    cbc[AUTH_VECTOR_LEN];
	char                    *end;

	/* make standard output line buffered */
        if (setvbuf(stdout, NULL, _IOLBF, 0) != 0)
                return 1;
		
	/* Fake: do we need this? */
	portno = 111;
	
 	if ( rad_auth_config(server, secretkey) < 0)
	{
	       fprintf(stderr, "%s: can't open configuration file.\n", argv[0]);
	       return -1;
	}
	
	gethostname(nasname, sizeof(nasname));
	nas_ipaddr = get_ipaddr(nasname);

	dict_init(NULL);

	/*
	 *	Open a connection to the server.
	 */
	svp = getservbyname ("radius", "udp");
	if (svp == (struct servent *) 0)
		svc_port = PW_AUTH_UDP_PORT;
	else
		svc_port = ntohs((u_short) svp->s_port);

	/* Get the IP address of the authentication server */
	if ((auth_ipaddr = get_ipaddr(server)) == 0) {
		fprintf(stderr, "Couldn't find host %s\n", server);
		exit(1);
	}

	sockfd = socket (AF_INET, SOCK_DGRAM, 0);
	if (sockfd < 0) {
		perror ("socket");
		exit(1);
	}

	sin = (struct sockaddr_in *) &salocal;
        memset (sin, 0, sizeof (salocal));
	sin->sin_family = AF_INET;
	sin->sin_addr.s_addr = INADDR_ANY;

	local_port = 1025;
	do {
		local_port++;
		sin->sin_port = htons((u_short)local_port);
	} while((bind(sockfd, &salocal, sizeof (struct sockaddr_in)) < 0) &&
						local_port < 64000);
	if (local_port >= 64000) {
		close(sockfd);
		perror ("bind");
		exit(1);
	}

 	while (fgets(authstring, MAXLINE, stdin) != NULL)
  	{	/* protect me form to long lines */
 		if (strlen(authstring)>MAXLINE)
  		{	printf("ERR\n");
  			continue;
  		}
 		/* Strip off the trailing newline */
 		if (strlen(authstring) > 0)
 		    authstring[strlen(authstring) - 1] = '\0';
 
 		/* Parse out the username and password */
 		ptr = authstring;
 		while (isspace(*ptr))
 		    ptr++;
 		if ((end = strchr(ptr, ' ')) == NULL)
 		{
 		    printf("ERR\n");/* No password */
 		    continue;
  		}
 		*end = '\0';
 		strncpy(username, ptr, MAXPWNAM);
 		ptr = end+1;
 		while (isspace(*ptr))
 		    ptr++;
 		strncpy(passwd, ptr, MAXPWNAM);
 

		/* I'll protect you to yourself here...
		 * remove this if you think you need it.
		 */
		if ( !strcmp(username, "root") )
		{	printf("ERR\n");
			continue;
		}

		/*
		 *	Build an authentication request
		 */
		auth = (AUTH_HDR *)send_buffer;
		auth->code = PW_AUTHENTICATION_REQUEST;
		auth->id =  getpid() % 256 ;
		random_vector(vector);
		memcpy(auth->vector, vector, AUTH_VECTOR_LEN);
		total_length = AUTH_HDR_LEN;
		ptr = auth->data;

		/*
		 *	User Name
		 */
		*ptr++ = PW_USER_NAME;
		length = strlen(username);
		if(length > MAXPWNAM) {
			length = MAXPWNAM;
		}
		*ptr++ = length + 2;
		memcpy(ptr, username, length);
		ptr += length;
		total_length += length + 2;

		/*
		 *	Password
		 */
		length = strlen(passwd);
		if(length > MAXPASS) {
			length = MAXPASS;
		}
		memset(passbuf, 0, MAXPASS);
		memcpy(passbuf, passwd, length);

 		/* 
 		 * Length is rounded up to multiple of 16,
 		 * and the password is encoded in blocks of 16 
 		 * with cipher block chaining
 		 */
 		length = ((length / AUTH_VECTOR_LEN)+1) * AUTH_VECTOR_LEN;
 
 		*ptr++ = PW_PASSWORD;
 		*ptr++ = length + 2;
 
  		secretlen = strlen(secretkey);
 		/* Set up the Cipher block chain */
 		memcpy(cbc, auth->vector, AUTH_VECTOR_LEN);
 		for (j =0; j < length; j+=AUTH_VECTOR_LEN)
 		{
 		    /* Calculate the MD5 Digest */
 		    strcpy(md5buf, secretkey);
 		    memcpy(md5buf + secretlen, cbc, AUTH_VECTOR_LEN);
 		    md5_calc(cbc, md5buf, secretlen + AUTH_VECTOR_LEN);
 		    
 		    /* Xor the password into the MD5 digest */
 		    for(i = 0;i < AUTH_VECTOR_LEN;i++) {
 			*ptr++ = (cbc[i] ^= passbuf[j+i]);
 		    }
  		}
 		total_length += length + 2;

		*ptr++ = PW_NAS_PORT_ID;
		*ptr++ = 6;
		
		ui=htonl(portno);
		memcpy(ptr, &ui, 4);
		ptr += 4;
		total_length += 6;

		*ptr++ = PW_NAS_IP_ADDRESS;
		*ptr++ = 6;
		
		ui=htonl(nas_ipaddr);
		memcpy(ptr, &ui, 4);
		ptr += 4;
		total_length += 6;

		auth->length = htonl(total_length);

 		sin = (struct sockaddr_in *) &saremote;
 		memset (sin, 0, sizeof (saremote));
 		sin->sin_family = AF_INET;
 		sin->sin_addr.s_addr = htonl(auth_ipaddr);
 		sin->sin_port = htons(svc_port);

		/*
		 *	Send the request we've built.
		 */
		sendto(sockfd, (char *)auth, total_length, 0,
			&saremote, sizeof(struct sockaddr_in));
 		tv.tv_sec = 3;
 		tv.tv_usec = 0;
 		FD_ZERO(&readfds);
 		FD_SET(sockfd, &readfds);
 		if (select(sockfd + 1, &readfds, NULL, NULL, &tv) == 0)
 			continue;
		salen = sizeof (saremote);
 		result = recvfrom (sockfd, recv_buffer, sizeof(i_recv_buffer),
  			0, &saremote, &salen);

		if ( result > 0 )
		{	if ( result_recv(sin->sin_addr.s_addr, sin->sin_port, recv_buffer, result) == 0)
				printf("OK\n");
			else
				printf("ERR\n");
		}
		else
			printf("ERR\n");
			
	}
	close(sockfd);
	exit(1);
}
